import { world, system, BlockVolume } from "@minecraft/server"; 
import { ActionFormData } from "@minecraft/server-ui";

world.getDimension('overworld').runCommand('gamerule showtags false')

const CONFIG = {
    typeId: 'wan:xray_goggles',
    blockRadius: 15
}

const UPGRADES = [
    'wan:coal_upgrade',
    'wan:copper_upgrade',
    'wan:iron_upgrade',
    'wan:gold_upgrade',
    'wan:redstone_upgrade',
    'wan:lapis_upgrade',
    'wan:diamond_upgrade',
    'wan:emerald_upgrade',
    'wan:debris_upgrade',
    'wan:quartz_upgrade',
    'wan:chest_upgrade',
    'wan:spawner_upgrade',
    'wan:shulker_upgrade'
]

const CATEGORIES = {
    "wan:coal_upgrade": {
        icon: 1,
        typeIds: ["coal_ore", "deepslate_coal_ore"],
        color: {r: 15, g: 15, b: 15}
    },
    "wan:copper_upgrade": {
        icon: 2,
        typeIds: ["copper_ore", "deepslate_copper_ore"],
        color: {r: 85, g: 35, b: 25}
    },
    "wan:iron_upgrade": {
        icon: 3,
        typeIds: ["iron_ore", "deepslate_iron_ore"],
        color: {r: 70, g: 60, b: 50}
    },
    "wan:gold_upgrade": {
        icon: 4,
        typeIds: ["gold_ore", "deepslate_gold_ore", "nether_gold_ore"],
        color: {r: 90, g: 50, b: 0}
    },
    "wan:redstone_upgrade": {
        icon: 5,
        typeIds: ["redstone_ore", "lit_redstone_ore", "deepslate_redstone_ore", "lit_deepslate_redstone_ore"],
        color: {r: 100, g: 0, b: 0}
    },
    "wan:lapis_upgrade": {
        icon: 6,
        typeIds: ["lapis_ore", "deepslate_lapis_ore"],
        color: {r: 15, g: 15, b: 95}
    },
    "wan:diamond_upgrade": {
        icon: 7,
        typeIds: ["diamond_ore", "deepslate_diamond_ore"],
        color: {r: 45, g: 95, b: 95}
    },
    "wan:emerald_upgrade": {
        icon: 8,
        typeIds: ["emerald_ore", "deepslate_emerald_ore"],
        color: {r: 10, g: 85, b: 15}
    },
    "wan:debris_upgrade": {
        icon: 9,
        typeIds: ["ancient_debris"],
        color: {r: 45, g: 35, b: 30}
    },
    "wan:quartz_upgrade": {
        icon: 10,
        typeIds: ["quartz_ore"],
        color: {r: 90, g: 85, b: 75}
    },
    "wan:chest_upgrade": {
        icon: 11,
        typeIds: ["chest", "trapped_chest", "barrel", "ender_chest"],
        color: {r: 90, g: 20, b: 0}
    },
    "wan:shulker_upgrade": {
        icon: 13,
        typeIds: [ "black_shulker_box", "blue_shulker_box", "brown_shulker_box", "cyan_shulker_box", "gray_shulker_box", "green_shulker_box", "light_blue_shulker_box", "light_gray_shulker_box", "lime_shulker_box", "magenta_shulker_box", "orange_shulker_box", "pink_shulker_box", "purple_shulker_box", "red_shulker_box", "undyed_shulker_box", "white_shulker_box", "yellow_shulker_box" ],
        color: {r: 90, g: 40, b: 70}
    },
    "wan:spawner_upgrade": {
        displayName: 'Spawners',
        icon: 12,
        typeIds: ["mob_spawner", "trial_spawner"],
        color: {r: 20, g: 15, b: 45}
    }
}

system.runInterval(() => {
    for (const player of world.getPlayers()) {
        const headItem = player.getComponent('equippable').getEquipment('Head')
        if ((!headItem || (headItem && headItem.typeId !== CONFIG.typeId))) {
            if (!player.getDynamicProperty('disabled')) {
                player.setDynamicProperty('disabled', 1)
                killAllOutlines(player) 
            }
            continue
        }
        if (player.getDynamicProperty('disabled') == 1) player.setDynamicProperty('disabled', 0)
        const foundCategories = Object.entries(CATEGORIES).filter(([key, category]) => headItem.getDynamicProperty(key) !== undefined)
        const blockCategories = foundCategories.map(entry => entry[1]).filter(category => category.typeIds)
        if (!blockCategories.length) continue
        const typeIds = blockCategories.flatMap((category) => category.typeIds)
        const locations = getBlockLocationsInRadius(player.location, player.dimension, CONFIG.blockRadius, typeIds)
        if (!locations) continue
        locations.forEach((location) => {
            const foundEntities = player.dimension.getEntitiesAtBlockLocation(location)
            const block = player.dimension.getBlock(location)
            const foundCategory = foundCategories.filter(([key, category]) => category.typeIds).find(([key, category]) => category.typeIds.find((typeId) => block?.typeId.includes(typeId)))
            if (!headItem.getDynamicProperty(foundCategory[0])) return
            if (!foundEntities.find(entity => entity?.typeId == 'wan:block_outline')) {
                const outline = player.dimension.spawnEntity('wan:block_outline', {x: location.x + 0.5, y: location.y + 0.5 , z: location.z + 0.5, })
                outline.setDynamicProperty('creator', player.id)
                outline.setProperty('wan:red', foundCategory[1].color.r)
                outline.setProperty('wan:green', foundCategory[1].color.g)
                outline.setProperty('wan:blue', foundCategory[1].color.b)
                outline.setProperty('wan:icon', foundCategory[1].icon)
            }
        })
    }
},10)
function getBlockLocationsInRadius(location, dimension, radius = 10, typeIds) {
    if (!location) return
    const startLocation = {
        x: location.x + radius,
        y: location.y + radius,
        z: location.z + radius,
    }
    const endLocation = {
        x: location.x - radius,
        y: location.y - radius,
        z: location.z - radius,
    }
    const volume = new BlockVolume(startLocation, endLocation)
    const blocks = dimension.getBlocks(volume, {includeTypes: typeIds}, true)
    const blockLocations = [...blocks.getBlockLocationIterator()]
    return blockLocations
}

function killAllOutlines(player) {
    const entities = player.dimension.getEntities({type: 'wan:block_outline', location: player.location})
    entities.forEach(entity => {
        if (entity.typeId == 'wan:block_outline' && entity.getDynamicProperty('creator') == player.id) {
            entity.triggerEvent('despawn')
        }
    })
}

world.beforeEvents.playerBreakBlock.subscribe(({player, block}) => {
    const location = block.location
    const hasEntity = player.dimension.getEntitiesAtBlockLocation(location)
    system.run(() => {
        if (!hasEntity.find(entity => entity?.typeId == 'wan:block_outline')) return
        hasEntity.forEach(entity => {
            if (entity.typeId !== 'wan:block_outline') return
            entity.triggerEvent('despawn')
        })
    })
})

world.afterEvents.blockExplode.subscribe((data) => {
    const hasEntity = data.dimension.getEntitiesAtBlockLocation(data.block.location)
    if (!hasEntity.find(entity => entity?.typeId == 'wan:block_outline')) return
    hasEntity.forEach(entity => {
        entity.triggerEvent('despawn')
    })
})

function killSpecificOutlines(player, typeIds) {
    const entities = player.dimension.getEntities({type: 'wan:block_outline', location: player.location})
    entities.forEach(entity => {
        const block = player.dimension.getBlock(entity.location)
        if (entity.getDynamicProperty('creator') !== player.id) return
        if (!typeIds.some(typeId => block?.typeId.includes(typeId))) return 
        entity.triggerEvent('despawn')
    })
}

function selectionMenu(player, goggles) {
    const f = new ActionFormData()
    const buttons = []
    Object.entries(CATEGORIES).forEach(([key, category]) => {
        if (goggles.getDynamicProperty(key) == undefined) return
        const isActive = goggles.getDynamicProperty(key)
        const status = isActive ? `§aEnabled` : `§cDisabled`
        buttons.push({text: `${loreMap(key)}\n${status}`, category, key})
    })
    if (!buttons.length) return error(player, '§7[§c!§7] Get an §cUpgrade to Open §7this §cMenu§7.')
    f.button('Help Menu')
    buttons.forEach((button) => {
        f.button(button.text)
    })
    f.show(player).then(({canceled, selection})=> {
        if (canceled) return
        if (selection == 0) return helpMenu(player, goggles)
        const isActive = goggles.getDynamicProperty(buttons[selection - 1].key)
        if (!isActive) goggles.setDynamicProperty(buttons[selection - 1].key, true)
        else {
            goggles.setDynamicProperty(buttons[selection - 1].key, false)
            if (buttons[selection - 1].category.typeIds) killSpecificOutlines(player, buttons[selection - 1].category.typeIds)
        }
        player.getComponent('inventory').container.setItem(player.selectedSlotIndex, goggles)
    })
}

function findItem(inventory, typeId) {
    for (let i = 0; i < inventory.size; i++) {
        const item = inventory.getItem(i)
        if (!item) continue
        if (item.typeId !== typeId) continue
        return {item: item, slot: i}
    }
}

world.afterEvents.itemUse.subscribe(({itemStack, source}) => {
    const equipment = source.getComponent('equippable')
    const inventory = source.getComponent('inventory').container
    const swappedItem = inventory.getItem(source.selectedSlotIndex)
    if (itemStack.typeId == CONFIG.typeId && source.isSneaking) {
        selectionMenu(source, itemStack)
        equipment.setEquipment('Head')
        if (swappedItem && source.getGameMode() !== 'creative') inventory.addItem(swappedItem)
        return inventory.setItem(source.selectedSlotIndex, itemStack)
    }
    if (!UPGRADES.includes(itemStack.typeId)) return
    let goggles = findItem(inventory, CONFIG.typeId)
    if (!goggles) {
        const headEquipment = equipment.getEquipment('Head')
        if (headEquipment && headEquipment.typeId == CONFIG.typeId) goggles = headEquipment
        else return error(source, '§7[§c!§7] §7You §cDon\'t Have Goggles §7in Your §cInventory§7.')
    }
    upgradeGoggles(source, itemStack, goggles.item ?? goggles, goggles.slot)
})

const loreMap = (typeId) => {
    switch (typeId) {
        case UPGRADES[0]: return ' §8Coal Upgrade'
        case UPGRADES[1]: return ' §nCopper Upgrade'
        case UPGRADES[2]: return ' §fIron Upgrade'
        case UPGRADES[3]: return ' §eGold Upgrade'
        case UPGRADES[4]: return ' §cRedstone Upgrade'
        case UPGRADES[5]: return ' §9Lapis Upgrade'
        case UPGRADES[6]: return ' §bDiamond Upgrade'
        case UPGRADES[7]: return ' §aEmerald Upgrade'
        case UPGRADES[8]: return ' §jDebris Upgrade'
        case UPGRADES[9]: return ' §fQuartz Upgrade'
        case UPGRADES[10]: return ' §6Chest Upgrade'
        case UPGRADES[11]: return ' §tSpawner Upgrade'
        case UPGRADES[12]: return ' §uShulker Upgrade'
    }
}

function upgradeGoggles(player, upgradeItem, gogglesItem, slot) {
    if (gogglesItem.getDynamicProperty(upgradeItem.typeId)) return error(player, '§7[§c!§7] Goggles §cAlready Contain §7This §cUpgrade§7.')
    gogglesItem.setDynamicProperty(upgradeItem.typeId, true)
    const lore = gogglesItem.getLore()
    lore[0] = '§r§5Contains:'
    lore.push('§r' + loreMap(upgradeItem.typeId))
    player.sendMessage('§7[§a!§7] Goggles §aSuccessfully Upgraded§7.')
    gogglesItem.setLore(lore)
    if (slot !== undefined) player.getComponent('inventory').container.setItem(slot, gogglesItem)
    else player.getComponent('equippable').setEquipment('Head', gogglesItem)
    if (upgradeItem.amount == 1) return player.getComponent('inventory').container.setItem(player.selectedSlotIndex)
    upgradeItem.amount = upgradeItem.amount - 1
    player.getComponent('inventory').container.setItem(player.selectedSlotIndex, upgradeItem)
}

function error(player, message, sound = 'note.bass') {
    player.sendMessage(message)
    player.playSound(sound)
}


function helpMenu(player, goggles) {
    const f = new ActionFormData()
    f.title('Help Menu')
    f.body(text.join('\n').replace('$player', player.name))
    f.button('Go Back')
    f.show(player).then((r) => {
        if (r.selection == 0) selectionMenu(player, goggles)
    })
}

const text = [
    "§7Hello §b$player, §7you have been selected to test our new creation! the §aX-Ray Goggles§7.",
    "",
    "§eHow do they work?",
    "§7It's easy, all you have to do is put them on your head and get the necessary upgrades to see the ores!",
    "",
    "§eWhat are upgrades?",
    "§7We have created §a13 upgrades §7for you! Upgrades are items that you can craft and apply to your §aX-Ray Goggles§7.",
    "",
    "§eHere's a list of the upgrades:",
    "    §8Coal Upgrade",
    "    §nCopper Upgrade",
    "    §fIron Upgrade",
    "    §eGold Upgrade",
    "    §cRedstone Upgrade",
    "    §9Lapis Upgrade",
    "    §bDiamond Upgrade",
    "    §aEmerald Upgrade",
    "    §jDebris Upgrade",
    "    §fQuartz Upgragrade",
    "    §6Chest Upgrade",
    "    §tSpawner Upgrade",
    "    §uShulker Upgrade",
    "",
    "§eHow can i Enable them?",
    "§7To §3enable §7them you will have to §3sneak and right-click §7the item while in your hand, that will open a menu in which you can enable/disable each upgrade",
    "",                                                                                                         
    "§8[§c!§8] §7For §eLow-end Devices§7, it's recommended to:",
    " »  §7Add upgrades of only §ccertain §7ores.",
    " »  §cDisable §7upgrades you do not need.",
    " »  §cDisable §nCopper §7and §8Coal."
]